VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsOCR"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' Autor: F. Schler (frank@activevb.de)
' Datum: 06/2022

Option Explicit

' Basiert auf:
' https://www.autohotkey.com/boards/viewtopic.php?f=6&t=72674

' ----==== Const ====----
Private Const S_OK As Long = &H0&
Private Const API_TRUE As Long = &H1&
Private Const CC_STDCALL As Long = &H4&
Private Const GdiPlusVersion As Long = &H1&

Private Const WindowsMediaOcrOcrEngine As String = "Windows.Media.Ocr.OcrEngine"
Private Const WindowsGraphicsImagingBitmapDecoder As String = "Windows.Graphics.Imaging.BitmapDecoder"

Private Const IID_IClosable As String = "{30d5a829-7fa4-4026-83bb-d75bae4ea99e}"
Private Const IID_IAsyncInfo As String = "{00000036-0000-0000-c000-000000000046}"
Private Const IID_IBitmapFrame As String = "{72a49a1c-8081-438d-91bc-94ecfc8185c6}"
Private Const IID_ILanguageFactory As String = "{9b0252ac-0c27-44f8-b792-9793fb66c63e}"
Private Const IID_IOcrEngineStatics As String = "{5bffa85a-3384-3540-9940-699120d428a8}"
Private Const IID_IRandomAccessStream As String = "{905a0fe1-bc53-11df-8c49-001e4fc686da}"
Private Const IID_IBitmapDecoderStatics As String = "{438ccb26-bcef-4e95-bad6-23a822e58d01}"
Private Const IID_IBitmapFrameWithSoftwareBitmap As String = "{fe287c9a-420c-4963-87ad-691436e08383}"

Private Const CLSID_GDIPPngEncoder As String = "{557cf406-1a04-11d3-9a73-0000f81ef32e}"

' ----==== Enums ====----
Private Enum vtb_Interfaces
    
    ' IUnknown
    IUnknown_QueryInterface = 0
    IUnknown_Release = 2

    ' IInspectable
    IInspectable_GetIids = 3
    IInspectable_GetRuntimeClassName = 4
    IInspectable_GetTrustLevel = 5

    'IReadOnlyList
    IReadOnlyList_GetAt = 6
    IReadOnlyList_GetSize = 7
    
    ' ILanguage
    ILanguage_GetLanguageTag = 6
    ILanguage_GetDisplayName = 7
    ILanguage_GetNativeName = 8
    ILanguage_GetScript = 9

    ' IOcrEngineStatics
    IOcrEngineStatics_GetMaxImageDimension = 6
    IOcrEngineStatics_GetAvailableRecognizerLanguages = 7
    IOcrEngineStatics_IsLanguageSupported = 8
    IOcrEngineStatics_TryCreateFromLanguage = 9
    IOcrEngineStatics_TryCreateFromUserProfileLanguages = 10
    
    ' IOcrEngine
    IOcrEngine_RecognizeAsync = 6
    IOcrEngine_GetRecognizerLanguage = 7
    
    ' IOcrResult
    IOcrResult_GetLines = 6
    IOcrResult_GetTextAngle = 7
    IOcrResult_GetText = 8
    
    ' IOcrLine
    IOcrLine_GetWords = 6
    IOcrLine_GetText = 7
    
    ' IOcrWord
    IOcrWord_GetBoundingRect = 6
    IOcrWord_GetText = 7
    
    ' IBitmapDecoderStatics
    IBitmapDecoderStatics_CreateAsync = 14
    
    ' IBitmapFrame
    IBitmapFrame_GetPixelWidth = 12
    IBitmapFrame_GetPixelHeight = 13

    ' IBitmapFrameWithSoftwareBitmap
    IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync = 6
    
    ' IReferenceDouble
    IReferenceDouble_GetValue = 6
    
    ' IAsyncInfo
    IAsyncInfo_GetStatus = 7
    IAsyncInfo_Close = 10

    ' IAsyncInterface
    IAsyncInterface_GetResults = 8
    
    ' IClosable
    IClosable_Close = 6

End Enum

Private Enum FileAccessMode
    Read = 0
    ReadWrite = 1
End Enum

Private Enum BSOS_OPTIONS
    BSOS_DEFAULT = 0
    BSOS_PREFERDESTINATIONSTREAM = 1
End Enum

Private Enum AsyncStatus
    Started = 0
    Completed = 1
    Canceled = 2
    Error = 3
End Enum

Private Enum TrustLevel
    BaseTrust = 0
    PartialTrust = 1
    FullTrust = 2
End Enum

Private Enum Status
    OK = 0
    GenericError = 1
    InvalidParameter = 2
    OutOfMemory = 3
    ObjectBusy = 4
    InsufficientBuffer = 5
    NotImplemented = 6
    Win32Error = 7
    WrongState = 8
    Aborted = 9
    FileNotFound = 10
    ValueOverflow = 11
    AccessDenied = 12
    UnknownImageFormat = 13
    FontFamilyNotFound = 14
    FontStyleNotFound = 15
    NotTrueTypeFont = 16
    UnsupportedGdiplusVersion = 17
    GdiplusNotInitialized = 18
    PropertyNotFound = 19
    PropertyNotSupported = 20
    ProfileNotFound = 21
End Enum

' ----==== Types ====----
Private Type GUID
    data1 As Long
    data2 As Long
    data3 As Long
    data4 As Long
End Type

Private Type GDIPlusStartupInput
    GdiPlusVersion As Long
    DebugEventCallback As Long
    SuppressBackgroundThread As Long
    SuppressExternalCodecs As Long
End Type

Private Type GdiplusStartupOutput
    NotificationHook As Long
    NotificationUnhook As Long
End Type

Private Type OcrLanguage
    LanguageTag As String
    DisplayName As String
    NativeName As String
    Script As String
End Type

Private Type OcrWord
    TextLineWord As String
    TextLineWordRectF As RECTF
End Type

Private Type OcrLine
    TextLine As String
    TextLineWordsCount As Long
    TextLineWords() As OcrWord
End Type

Private Type OcrResults
    TextOcrLanguage As OcrLanguage
    TextAngle As Double
    TextComplete As String
    TextLineCount As Long
    TextLines() As OcrLine
End Type

' ----==== Ole32.dll Declarations ====----
Private Declare Function CLSIDFromString Lib "Ole32.dll" ( _
                         ByVal pstring As Long, _
                         ByRef pCLSID As GUID) As Long
                         
Private Declare Function StringFromGUID Lib "Ole32.dll" _
                         Alias "StringFromGUID2" ( _
                         ByRef rguid As GUID, _
                         ByVal lpsz As Long, _
                         ByVal cchMax As Long) As Long
                         
Private Declare Sub CoTaskMemFree Lib "Ole32.dll" ( _
                    ByVal hMem As Long)
                         
' ----==== Oleaut32.dll Declarations ====----
Private Declare Function DispCallFunc Lib "Oleaut32.dll" ( _
                         ByVal pvInstance As Long, _
                         ByVal oVft As Long, _
                         ByVal cc As Long, _
                         ByVal vtReturn As VbVarType, _
                         ByVal cActuals As Long, _
                         ByVal prgvt As Any, _
                         ByVal prgpvarg As Any, _
                         ByRef pvargResult As Variant) As Long

' ----==== Kernel32.dll Declarations ====----
Private Declare Sub CopyMemory Lib "Kernel32.dll" _
                    Alias "RtlMoveMemory" ( _
                    ByRef pTo As Any, _
                    ByRef uFrom As Any, _
                    ByVal lSize As Long)

Private Declare Function lstrlenW Lib "Kernel32.dll" ( _
                         ByVal lpString As Long) As Long

' ----==== Combase.dll Declarations ====----
Private Declare Function WindowsCreateString Lib "Combase.dll" ( _
                         ByVal sourceString As Long, _
                         ByVal lenght As Long, _
                         ByRef hString As Long) As Long

Private Declare Function WindowsDeleteString Lib "Combase.dll" ( _
                         ByVal sourceString As Long) As Long

Private Declare Function WindowsGetStringRawBuffer Lib "Combase.dll" ( _
                         ByVal sourceString As Long, _
                         ByRef length As Long) As Long

Private Declare Function RoGetActivationFactory Lib "Combase.dll" ( _
                         ByVal activatableClassId As Long, _
                         ByRef riid As GUID, _
                         ByRef factory As Long) As Long

' ----==== Shlwapi.dll Declarations ====----
Private Declare Function SHCreateMemStream Lib "Shlwapi.dll" ( _
                         ByVal pInit As Long, _
                         ByVal cbInit As Long) As Long

' ----==== ShCore.dll Declarations ====----
Private Declare Function CreateRandomAccessStreamOnFile Lib "ShCore.dll" ( _
                         ByVal filePath As Long, _
                         ByVal accessMode As FileAccessMode, _
                         ByRef riid As GUID, _
                         ByRef ppv As Long) As Long

Private Declare Function CreateRandomAccessStreamOverStream Lib "ShCore.dll" ( _
                         ByVal pStream As Long, _
                         ByVal options As BSOS_OPTIONS, _
                         ByRef riid As GUID, _
                         ByRef ppv As Long) As Long

' ----==== Gdiplus.dll Declarations ====----
Private Declare Function GdiplusShutdown Lib "GdiPlus.dll" ( _
                         ByVal token As Long) As Status
                         
Private Declare Function GdiplusStartup Lib "GdiPlus.dll" ( _
                         ByRef token As Long, _
                         ByRef lpInput As GDIPlusStartupInput, _
                         ByRef lpOutput As GdiplusStartupOutput) As Status

Private Declare Function GdipCreateBitmapFromHBITMAP Lib "GdiPlus.dll" ( _
                         ByVal hbm As Long, _
                         ByVal hpal As Long, _
                         ByRef bitmap As Long) As Status

Private Declare Function GdipDisposeImage Lib "GdiPlus.dll" ( _
                         ByVal image As Long) As Status

Private Declare Function GdipSaveImageToStream Lib "GdiPlus.dll" ( _
                         ByVal mImage As Long, _
                         ByVal mStream As Long, _
                         ByRef mClsidEncoder As GUID, _
                         ByVal mEncoderParams As Long) As Status
                         
' ----==== Variablen ====----
Private m_lngGdipToken As Long
Private m_bolIsOcrInitialized As Boolean
Private m_lngMaxImageDimension As Long
Private m_lngOcrTextLineCount As Long
Private m_lngAvailableOcrLanguagesCount As Long
Private m_pIOcrEngine As Long
Private m_pILanguageFactory As Long
Private m_pIOcrEngineStatics As Long
Private m_pIBitmapDecoderStatics As Long
Private m_tAvailableOcrLanguages() As OcrLanguage
Private m_tOcrResults As OcrResults

Private Sub Class_Initialize()
    
    m_bolIsOcrInitialized = False
    m_lngMaxImageDimension = 0&
    m_lngOcrTextLineCount = 0&
    m_lngAvailableOcrLanguagesCount = 0&
    m_pIOcrEngine = 0&
    m_pILanguageFactory = 0&
    m_pIOcrEngineStatics = 0&
    m_pIBitmapDecoderStatics = 0&

    If GetActivationFactory(WindowsMediaOcrOcrEngine, _
                            IID_IOcrEngineStatics, _
                            m_pIOcrEngineStatics) Then
        If GetActivationFactory(WindowsGraphicsImagingBitmapDecoder, _
                                IID_IBitmapDecoderStatics, _
                                m_pIBitmapDecoderStatics) Then
                                    
            Dim tGdipStartupInput As GDIPlusStartupInput
            Dim tGdipStartupOutput As GdiplusStartupOutput
            tGdipStartupInput.GdiPlusVersion = GdiPlusVersion
            If GdiplusStartup(m_lngGdipToken, tGdipStartupInput, _
                              tGdipStartupOutput) = OK Then
                m_bolIsOcrInitialized = True
            End If
                                    
            If m_bolIsOcrInitialized Then
                Call Invoke(m_pIOcrEngineStatics, _
                            IOcrEngineStatics_GetMaxImageDimension, _
                            VarPtr(m_lngMaxImageDimension))
                m_lngAvailableOcrLanguagesCount = GetAvailableOcrLanguages
            End If
        End If
    End If
End Sub

Private Sub Class_Terminate()
    Call Release(m_pIOcrEngine)
    Call Release(m_pIBitmapDecoderStatics)
    Call Release(m_pIOcrEngineStatics)
    If m_lngGdipToken <> 0& Then Call GdiplusShutdown(m_lngGdipToken)
End Sub

' ----==== Properties ====----
Public Property Get GetIsOcrInitialized() As Boolean
    GetIsOcrInitialized = m_bolIsOcrInitialized
End Property

Public Property Get GetOcrResultUsedLanguageTag() As String
    GetOcrResultUsedLanguageTag = _
        m_tOcrResults.TextOcrLanguage.LanguageTag
End Property

Public Property Get GetOcrResultUsedLanguageDisplayName() As String
    GetOcrResultUsedLanguageDisplayName = _
        m_tOcrResults.TextOcrLanguage.DisplayName
End Property

Public Property Get GetOcrResultUsedLanguageNativeName() As String
    GetOcrResultUsedLanguageNativeName = _
        m_tOcrResults.TextOcrLanguage.NativeName
End Property

Public Property Get GetOcrResultUsedLanguageScript() As String
    GetOcrResultUsedLanguageScript = _
        m_tOcrResults.TextOcrLanguage.Script
End Property

Public Property Get GetOcrResultTextAngle() As Double
    GetOcrResultTextAngle = m_tOcrResults.TextAngle
End Property

Public Property Get GetOcrResultTextComplete() As String
    GetOcrResultTextComplete = m_tOcrResults.TextComplete
End Property

Public Property Get GetOcrResultTextCompleteWithLineBreak() As String
    Dim strRet As String
    Dim lngCount As Long
    lngCount = Me.GetOcrResultTextLineCount
    If lngCount > 0 Then
        Dim lngIndex As Long
        For lngIndex = 0 To lngCount - 1
            strRet = strRet & Me.GetOcrResultTextLineFromIndex(lngIndex)
            If lngIndex < lngCount - 1 Then
                strRet = strRet & vbNewLine
            End If
        Next
    End If
    GetOcrResultTextCompleteWithLineBreak = strRet
End Property

Public Property Get GetOcrResultTextLineCount() As Long
    GetOcrResultTextLineCount = m_tOcrResults.TextLineCount
End Property

Public Property Get GetOcrResultTextLineFromIndex(ByVal lngLineIndex As Long) As String
    GetOcrResultTextLineFromIndex = _
        m_tOcrResults.TextLines(lngLineIndex).TextLine
End Property

Public Property Get GetOcrResultTextLineWordsCountFromIndex(ByVal lngLineIndex As Long) As Long
    GetOcrResultTextLineWordsCountFromIndex = _
        m_tOcrResults.TextLines(lngLineIndex).TextLineWordsCount
End Property

Public Property Get GetOcrResultTextLineWordFromIndex(ByVal lngLineIndex As Long, _
                                                      ByVal lngWordIndex As Long) As String
    GetOcrResultTextLineWordFromIndex = _
        m_tOcrResults.TextLines(lngLineIndex).TextLineWords(lngWordIndex).TextLineWord
End Property

Friend Property Get GetOcrResultTextLineWordRectFromIndex(ByVal lngLineIndex As Long, _
                                                          ByVal lngWordIndex As Long) As RECTF
    GetOcrResultTextLineWordRectFromIndex = _
        m_tOcrResults.TextLines(lngLineIndex).TextLineWords(lngWordIndex).TextLineWordRectF
End Property

Public Property Get GetMaxImageDimensionForOcr() As Long
    GetMaxImageDimensionForOcr = m_lngMaxImageDimension
End Property

Public Property Get GetAvailableOcrLanguagesCount() As Long
    GetAvailableOcrLanguagesCount = _
        m_lngAvailableOcrLanguagesCount
End Property

Public Property Get GetAvailableOcrLanguageTag(ByVal lngIndex As Long) As String
    GetAvailableOcrLanguageTag = _
        m_tAvailableOcrLanguages(lngIndex).LanguageTag
End Property

Public Property Get GetAvailableOcrLanguageDisplayName(ByVal lngIndex As Long) As String
    GetAvailableOcrLanguageDisplayName = _
        m_tAvailableOcrLanguages(lngIndex).DisplayName
End Property

Public Property Get GetAvailableOcrLanguageNativeName(ByVal lngIndex As Long) As String
    GetAvailableOcrLanguageNativeName = _
        m_tAvailableOcrLanguages(lngIndex).NativeName
End Property

Public Property Get GetAvailableOcrLanguageScript(ByVal lngIndex As Long) As String
    GetAvailableOcrLanguageScript = _
        m_tAvailableOcrLanguages(lngIndex).Script
End Property

Public Property Get InitializeOcrEngineWithLanguageIndex(ByVal lngLanguageIndex As Long) As Boolean
    InitializeOcrEngineWithLanguageIndex = _
        InitOcrEngineWithLanguageIndex(lngLanguageIndex)
End Property


' ----==== Subs & Functions ====----
Public Function GetOcrResultsFromHBitmap(ByVal hBitmap As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsOcrInitialized Then
        If m_pIOcrEngine <> 0& And hBitmap <> 0& Then
            Call GetUsedOcrLanguage
            Dim pBitmap As Long
            If GdipCreateBitmapFromHBITMAP(hBitmap, 0&, pBitmap) = OK Then
                Dim pIStream As Long
                pIStream = SHCreateMemStream(0&, 0&)
                If pIStream <> 0& Then
                    If GdipSaveImageToStream(pBitmap, pIStream, _
                                             Str2Guid(CLSID_GDIPPngEncoder), 0&) = OK Then
                        Dim pIRandomAccessStream As Long
                        If CreateRandomAccessStreamOverStream(pIStream, BSOS_DEFAULT, _
                                                              Str2Guid(IID_IRandomAccessStream), _
                                                              pIRandomAccessStream) = S_OK Then
                            bolRet = GetOcrResults(pIRandomAccessStream)
                            Call CloseAndRelease(pIRandomAccessStream)
                        End If
                    End If
                    Call Release(pIStream)
                End If
                Call GdipDisposeImage(pBitmap)
            End If
        End If
    End If
    GetOcrResultsFromHBitmap = bolRet
End Function

Public Function GetOcrResultsFromFile(ByVal strFile As String) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsOcrInitialized Then
        If m_pIOcrEngine <> 0& And Len(strFile) > 0& Then
            Call GetUsedOcrLanguage
            Dim pIRandomAccessStream As Long
            If CreateRandomAccessStreamOnFile(StrPtr(strFile), Read, _
                                              Str2Guid(IID_IRandomAccessStream), _
                                              pIRandomAccessStream) = S_OK Then
                bolRet = GetOcrResults(pIRandomAccessStream)
                Call CloseAndRelease(pIRandomAccessStream)
            End If
        End If
    End If
    GetOcrResultsFromFile = bolRet
End Function

Private Function GetOcrResults(ByVal pIRandomAccessStream As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    Dim pISoftwareBitmap As Long
    If GetISoftwareBitmapFromIRandomAccessStream(pIRandomAccessStream, _
                                                 pISoftwareBitmap) Then
        Dim pIOcrResult As Long
        If Invoke(m_pIOcrEngine, IOcrEngine_RecognizeAsync, _
                  pISoftwareBitmap, VarPtr(pIOcrResult)) = S_OK Then
            If WaitForAsync(pIOcrResult) Then
                m_tOcrResults.TextAngle = 0
                Dim pIReferenceDouble As Long
                If Invoke(pIOcrResult, IOcrResult_GetTextAngle, _
                          VarPtr(pIReferenceDouble)) = S_OK Then
                    Call Invoke(pIReferenceDouble, _
                                IReferenceDouble_GetValue, _
                                VarPtr(m_tOcrResults.TextAngle))
                    Call Release(pIReferenceDouble)
                End If
                m_tOcrResults.TextComplete = vbNullString
                Dim hStringComplete As Long
                If Invoke(pIOcrResult, IOcrResult_GetText, _
                          VarPtr(hStringComplete)) = S_OK Then
                    m_tOcrResults.TextComplete = _
                        Ptr2Str(WindowsGetStringRawBuffer(hStringComplete, 0&))
                End If
                bolRet = GetTextLinesFromOcrResult(pIOcrResult)
            End If
            Call Release(pIOcrResult)
        End If
        Call CloseAndRelease(pISoftwareBitmap)
    End If
    GetOcrResults = bolRet
End Function

Private Function GetTextLinesFromOcrResult(ByVal pIOcrResult As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    Dim pIReadOnlyList As Long
    If Invoke(pIOcrResult, IOcrResult_GetLines, _
              VarPtr(pIReadOnlyList)) = S_OK Then
        m_tOcrResults.TextLineCount = 0
        Dim tOcrLine() As OcrLine
        If Invoke(pIReadOnlyList, IReadOnlyList_GetSize, _
                  VarPtr(m_tOcrResults.TextLineCount)) = S_OK Then
            If m_tOcrResults.TextLineCount > 0 Then
                ReDim tOcrLine(m_tOcrResults.TextLineCount - 1)
                Dim lngOcrTextLineIndex As Long
                For lngOcrTextLineIndex = 0 To m_tOcrResults.TextLineCount - 1
                    Dim pIOcrLine As Long
                    If Invoke(pIReadOnlyList, IReadOnlyList_GetAt, _
                              lngOcrTextLineIndex, VarPtr(pIOcrLine)) = S_OK Then
                        Dim hString As Long
                        If Invoke(pIOcrLine, IOcrLine_GetText, _
                                  VarPtr(hString)) = S_OK Then
                            tOcrLine(lngOcrTextLineIndex).TextLine = Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                            Call GetWordsAndPositionFromTextLine(pIOcrLine, tOcrLine(lngOcrTextLineIndex))
                        End If
                        Call Release(pIOcrLine)
                    End If
                Next
                m_tOcrResults.TextLines = tOcrLine
                bolRet = True
            End If
        End If
        Call Release(pIReadOnlyList)
    End If
    GetTextLinesFromOcrResult = bolRet
End Function

Private Sub GetWordsAndPositionFromTextLine(ByVal pIOcrLine As Long, _
                                            ByRef tOcrLine As OcrLine)
    tOcrLine.TextLineWordsCount = 0
    Dim tOcrWord() As OcrWord
    Dim pIReadOnlyList As Long
    If Invoke(pIOcrLine, IOcrLine_GetWords, _
              VarPtr(pIReadOnlyList)) = S_OK Then
        If Invoke(pIReadOnlyList, IReadOnlyList_GetSize, _
                  VarPtr(tOcrLine.TextLineWordsCount)) = S_OK Then
            If tOcrLine.TextLineWordsCount > 0 Then
                ReDim tOcrWord(tOcrLine.TextLineWordsCount - 1)
                Dim lngLineWordsIndex As Long
                For lngLineWordsIndex = 0 To tOcrLine.TextLineWordsCount - 1
                    Dim pIOcrWord As Long
                    If Invoke(pIReadOnlyList, IReadOnlyList_GetAt, _
                              lngLineWordsIndex, VarPtr(pIOcrWord)) = S_OK Then
                        Dim tRectF As RECTF
                        If Invoke(pIOcrWord, IOcrWord_GetBoundingRect, _
                                  VarPtr(tRectF)) = S_OK Then
                            tOcrWord(lngLineWordsIndex).TextLineWordRectF = tRectF
                            Dim hString As Long
                            If Invoke(pIOcrWord, IOcrWord_GetText, _
                                      VarPtr(hString)) = S_OK Then
                                tOcrWord(lngLineWordsIndex).TextLineWord = _
                                    Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                            End If
                        End If
                        Call Release(pIOcrWord)
                    End If
                Next
                tOcrLine.TextLineWords = tOcrWord
            End If
        End If
        Call Release(pIReadOnlyList)
    End If
End Sub

Private Sub GetUsedOcrLanguage()
    If m_bolIsOcrInitialized Then
        If m_pIOcrEngine <> 0& Then
            Dim pILanguage As Long
            If Invoke(m_pIOcrEngine, IOcrEngine_GetRecognizerLanguage, _
                      VarPtr(pILanguage)) = S_OK Then
                Dim hString As Long
                If Invoke(pILanguage, ILanguage_GetLanguageTag, _
                          VarPtr(hString)) = S_OK Then
                    m_tOcrResults.TextOcrLanguage.LanguageTag = _
                        Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                End If
                If Invoke(pILanguage, ILanguage_GetDisplayName, _
                          VarPtr(hString)) = S_OK Then
                    m_tOcrResults.TextOcrLanguage.DisplayName = _
                        Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                End If
                If Invoke(pILanguage, ILanguage_GetNativeName, _
                          VarPtr(hString)) = S_OK Then
                    m_tOcrResults.TextOcrLanguage.NativeName = _
                        Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                End If
                If Invoke(pILanguage, ILanguage_GetScript, _
                          VarPtr(hString)) = S_OK Then
                    m_tOcrResults.TextOcrLanguage.Script = _
                        Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                End If
                Call Release(pILanguage)
            End If
        End If
    End If
End Sub

Private Function GetISoftwareBitmapFromIRandomAccessStream(ByVal pIRandomAccessStream As Long, _
                                                           ByRef pISoftwareBitmap As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsOcrInitialized Then
        Dim pIBitmapDecoder As Long
        If Invoke(m_pIBitmapDecoderStatics, IBitmapDecoderStatics_CreateAsync, _
                  pIRandomAccessStream, VarPtr(pIBitmapDecoder)) = S_OK Then
            If WaitForAsync(pIBitmapDecoder) Then
                Dim pIBitmapFrame As Long
                If Invoke(pIBitmapDecoder, IUnknown_QueryInterface, _
                          VarPtr(Str2Guid(IID_IBitmapFrame)), VarPtr(pIBitmapFrame)) = S_OK Then
                    Dim lngFrameWidth As Long
                    Dim lngFrameHeight As Long
                    Call Invoke(pIBitmapFrame, IBitmapFrame_GetPixelWidth, VarPtr(lngFrameWidth))
                    Call Invoke(pIBitmapFrame, IBitmapFrame_GetPixelHeight, VarPtr(lngFrameHeight))
                    If lngFrameWidth > m_lngMaxImageDimension Or _
                        lngFrameHeight > m_lngMaxImageDimension Then
                        Call MsgBox("Image is to big. Maximum: " & CStr(m_lngMaxImageDimension) & "x" & _
                                    CStr(m_lngMaxImageDimension) & " Pixel.", vbOKOnly Or vbInformation, "OCR")
                    Else
                        Dim pIBitmapFrameWithSoftwareBitmap As Long
                        If Invoke(pIBitmapDecoder, IUnknown_QueryInterface, _
                                  VarPtr(Str2Guid(IID_IBitmapFrameWithSoftwareBitmap)), _
                                  VarPtr(pIBitmapFrameWithSoftwareBitmap)) = S_OK Then
                            If Invoke(pIBitmapFrameWithSoftwareBitmap, _
                                      IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync, _
                                      VarPtr(pISoftwareBitmap)) = S_OK Then
                                If WaitForAsync(pISoftwareBitmap) Then
                                    bolRet = True
                                End If
                            End If
                            Call Release(pIBitmapFrameWithSoftwareBitmap)
                        End If
                    End If
                    Call Release(pIBitmapFrame)
                End If
            End If
            Call Release(pIBitmapDecoder)
        End If
    End If
    GetISoftwareBitmapFromIRandomAccessStream = bolRet
End Function

Private Function InitOcrEngineWithLanguageIndex(ByVal lngLanguageIndex As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsOcrInitialized Then
        Call Release(m_pIOcrEngine)
        If lngLanguageIndex >= 0 And lngLanguageIndex < m_lngAvailableOcrLanguagesCount Then
            Dim pILanguage As Long
            If GetILanguageByLanguageIndex(lngLanguageIndex, pILanguage) Then
                Dim bolIsLanguageSupported As Boolean
                If Invoke(m_pIOcrEngineStatics, IOcrEngineStatics_IsLanguageSupported, _
                          pILanguage, VarPtr(bolIsLanguageSupported)) = S_OK Then
                    If bolIsLanguageSupported Then
                        If Invoke(m_pIOcrEngineStatics, IOcrEngineStatics_TryCreateFromLanguage, _
                                  pILanguage, VarPtr(m_pIOcrEngine)) = S_OK Then
                            bolRet = True
                        End If
                    End If
                End If
                Call Release(pILanguage)
            End If
        End If
    End If
    InitOcrEngineWithLanguageIndex = bolRet
End Function

Private Function GetAvailableOcrLanguages() As Long
    Dim lngRet As Long
    If m_bolIsOcrInitialized Then
        Dim pIReadOnlyList As Long
        If Invoke(m_pIOcrEngineStatics, IOcrEngineStatics_GetAvailableRecognizerLanguages, _
                  VarPtr(pIReadOnlyList)) = S_OK Then
            Dim lngCount As Long
            ReDim m_tAvailableOcrLanguages(0)
            If Invoke(pIReadOnlyList, IReadOnlyList_GetSize, _
                      VarPtr(lngCount)) = S_OK Then
                If lngCount > 0 Then
                    ReDim m_tAvailableOcrLanguages(lngCount - 1)
                    Dim lngIndex As Long
                    For lngIndex = 0 To lngCount - 1
                        Dim pILanguage As Long
                        If Invoke(pIReadOnlyList, IReadOnlyList_GetAt, _
                                  lngIndex, VarPtr(pILanguage)) = S_OK Then
                            Dim hString As Long
                            If Invoke(pILanguage, ILanguage_GetLanguageTag, _
                                      VarPtr(hString)) = S_OK Then
                                m_tAvailableOcrLanguages(lngIndex).LanguageTag = _
                                    Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                            End If
                            If Invoke(pILanguage, ILanguage_GetDisplayName, _
                                      VarPtr(hString)) = S_OK Then
                                m_tAvailableOcrLanguages(lngIndex).DisplayName = _
                                    Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                            End If
                            If Invoke(pILanguage, ILanguage_GetNativeName, _
                                      VarPtr(hString)) = S_OK Then
                                m_tAvailableOcrLanguages(lngIndex).NativeName = _
                                    Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                            End If
                            If Invoke(pILanguage, ILanguage_GetScript, _
                                      VarPtr(hString)) = S_OK Then
                                m_tAvailableOcrLanguages(lngIndex).Script = _
                                    Ptr2Str(WindowsGetStringRawBuffer(hString, 0&))
                            End If
                            lngRet = lngRet + 1
                            Call Release(pILanguage)
                        End If
                    Next
                End If
            End If
            Call Release(pIReadOnlyList)
        End If
    End If
    GetAvailableOcrLanguages = lngRet
End Function

Private Function GetILanguageByLanguageIndex(ByVal lngLanguageIndex As Long, _
                                             ByRef pILanguage As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsOcrInitialized Then
        If lngLanguageIndex >= 0 And lngLanguageIndex < m_lngAvailableOcrLanguagesCount Then
            Dim pIReadOnlyList As Long
            If Invoke(m_pIOcrEngineStatics, IOcrEngineStatics_GetAvailableRecognizerLanguages, _
                      VarPtr(pIReadOnlyList)) = S_OK Then
                Dim lngCount As Long
                If Invoke(pIReadOnlyList, IReadOnlyList_GetSize, _
                          VarPtr(lngCount)) = S_OK Then
                    If lngCount > 0 Then
                        Dim lngIndex As Long
                        For lngIndex = 0 To lngCount - 1
                            If Invoke(pIReadOnlyList, IReadOnlyList_GetAt, _
                                      lngIndex, VarPtr(pILanguage)) = S_OK Then
                                If lngIndex = lngLanguageIndex Then
                                    bolRet = True
                                    Exit For
                                Else
                                    Call Release(pILanguage)
                                End If
                            End If
                        Next
                    End If
                End If
                Call Release(pIReadOnlyList)
            End If
        End If
    End If
    GetILanguageByLanguageIndex = bolRet
End Function

Private Function GetActivationFactory(ByVal ClassName As String, _
                                      ByVal IID As String, _
                                      ByRef pFactory As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    Dim hString As Long
    If WindowsCreateString(StrPtr(ClassName), Len(ClassName), hString) = S_OK Then
        If hString <> 0& Then
            If RoGetActivationFactory(hString, Str2Guid(IID), pFactory) = S_OK Then bolRet = True
            Call WindowsDeleteString(hString)
        End If
    End If
    GetActivationFactory = bolRet
End Function

Private Function WaitForAsync(ByRef pInterface As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If pInterface <> 0& Then
        Dim pIAsyncInfo As Long
        If Invoke(pInterface, IUnknown_QueryInterface, _
                  VarPtr(Str2Guid(IID_IAsyncInfo)), _
                  VarPtr(pIAsyncInfo)) = S_OK Then
            Dim eStatus As AsyncStatus
            Do
                If Invoke(pIAsyncInfo, IAsyncInfo_GetStatus, _
                          VarPtr(eStatus)) = S_OK Then DoEvents
            Loop While eStatus = Started
            If eStatus = Completed Then
                bolRet = True
            Else
                Call Invoke(pIAsyncInfo, IAsyncInfo_Close)
            End If
            Call Release(pIAsyncInfo)
            If bolRet Then
                Dim pRet As Long
                If Invoke(pInterface, IAsyncInterface_GetResults, VarPtr(pRet)) = S_OK Then
                    Call Release(pInterface)
                    pInterface = pRet
                End If
            Else
                Call Release(pInterface)
            End If
        End If
    End If
    WaitForAsync = bolRet
End Function

Private Function CloseAndRelease(ByRef pInterface As Long) As Long
    Dim lngRet As Long
    If pInterface <> 0& Then
        Dim pIClosable As Long
        If Invoke(pInterface, IUnknown_QueryInterface, _
                  VarPtr(Str2Guid(IID_IClosable)), _
                  VarPtr(pIClosable)) = S_OK Then
            Call Invoke(pIClosable, IClosable_Close)
            Call Release(pIClosable)
            Call Release(pInterface)
            pInterface = 0&
        End If
    End If
    CloseAndRelease = lngRet
End Function

Private Sub Release(ByRef pInterface As Long)
    Dim lngRet As Long
    If pInterface <> 0& Then
        Call Invoke(pInterface, IUnknown_Release)
        pInterface = 0&
    End If
End Sub

Private Sub GetInspectableInfo(ByVal pInspectableInterface As Long, _
                               ByVal strInterface As String)
    If pInspectableInterface <> 0& Then
        Debug.Print "Inspectable Info for: " & strInterface
        Debug.Print "RuntimeClassName = " & GetInspectable_GetRuntimeClassName(pInspectableInterface)
        Dim eTrustLevel As TrustLevel
        eTrustLevel = GetInspectable_GetTrustLevel(pInspectableInterface)
        Select Case eTrustLevel
            Case TrustLevel.BaseTrust
                Debug.Print "TrustLevel = BaseTrust"
            Case TrustLevel.FullTrust
                Debug.Print "TrustLevel = FullTrust"
            Case TrustLevel.PartialTrust
                Debug.Print "TrustLevel = PartialTrust"
            Case Else
                Debug.Print "TrustLevel = UnknownTrust"
        End Select
        Dim tIIDs() As GUID
        Dim lngIIDsCount As Long
        lngIIDsCount = GetInspectable_GetIIDs(pInspectableInterface, tIIDs)
        Debug.Print "AviableIIDsCount = " & CStr(lngIIDsCount)
        If lngIIDsCount > 0 Then
            Dim lngIIDsItem As Long
            For lngIIDsItem = 0 To lngIIDsCount - 1
                Debug.Print "    AviableIID " & CStr(lngIIDsItem + 1) & " = " & Guid2Str(tIIDs(lngIIDsItem))
            Next
        End If
        Debug.Print String(60, "-")
    End If
End Sub

Private Function GetInspectable_GetIIDs(ByVal pInterface As Long, _
                                        ByRef tIIDs() As GUID) As Long
    Dim lngIIDCount As Long
    If pInterface <> 0& Then
        Dim pIIDs As Long
        If Invoke(pInterface, IInspectable_GetIids, _
                  VarPtr(lngIIDCount), VarPtr(pIIDs)) = S_OK Then
            If lngIIDCount > 0 Then
                Dim lngBytes As Long
                Dim lngIIDItem As Long
                ReDim tIIDs(lngIIDCount - 1)
                For lngIIDItem = 0 To lngIIDCount - 1
                    lngBytes = Len(tIIDs(lngIIDItem))
                    Call CopyMemory(tIIDs(lngIIDItem), _
                                    ByVal pIIDs + (lngBytes * lngIIDItem), _
                                    lngBytes)
                Next
            End If
        End If
    End If
    GetInspectable_GetIIDs = lngIIDCount
End Function

Private Function GetInspectable_GetRuntimeClassName(ByVal pInterface As Long) As String
    Dim strRet As String
    If pInterface <> 0& Then
        Dim pStr As Long
        If Invoke(pInterface, IInspectable_GetRuntimeClassName, _
                  VarPtr(pStr)) = S_OK Then
            If pStr <> 0& Then strRet = Ptr2Str(pStr)
        End If
    End If
    GetInspectable_GetRuntimeClassName = strRet
End Function

Private Function GetInspectable_GetTrustLevel(ByVal pInterface As Long) As TrustLevel
    Dim eRet As TrustLevel
    If pInterface <> 0& Then
        Dim lngValue As Long
        If Invoke(pInterface, IInspectable_GetTrustLevel, _
                  VarPtr(lngValue)) = S_OK Then
            eRet = lngValue
        End If
    End If
    GetInspectable_GetTrustLevel = eRet
End Function


Private Function Invoke(ByVal pInterface As Long, _
                        ByVal vtb As vtb_Interfaces, _
                        ParamArray var()) As Variant
    If pInterface <> 0& Then
        Invoke = OleInvoke(pInterface, vtb, var)
    End If
End Function

Private Function OleInvoke(ByVal pInterface As Long, _
                           ByVal lngCmd As Long, _
                           ParamArray aParam()) As Variant
    If pInterface <> 0& Then
        If UBound(aParam) >= 0& Then
            Dim varParam As Variant
            varParam = aParam
            If IsArray(varParam) Then varParam = varParam(0)
            Dim lngCount As Long
            lngCount = UBound(varParam)
            Dim lngItem As Long
            Dim olePtr(10) As Long
            Dim oleTyp(10) As Integer
            For lngItem = 0& To lngCount
                oleTyp(lngItem) = VarType(varParam(lngItem))
                olePtr(lngItem) = VarPtr(varParam(lngItem))
            Next
        End If
        Dim varRet As Variant
        If DispCallFunc(pInterface, lngCmd * 4, CC_STDCALL, _
                        vbLong, lngItem, VarPtr(oleTyp(0)), _
                        VarPtr(olePtr(0)), varRet) <> S_OK Then
            Debug.Print "Fehler beim Aufrufen der Interface-Funktion!"
        End If
'        If varRet <> S_OK Then
'            Debug.Print "0x" & Hex$(varRet)
'        End If
    End If
    OleInvoke = varRet
End Function

Private Function Str2Guid(ByVal strGUID As String) As GUID
    Call CLSIDFromString(StrPtr(strGUID), Str2Guid)
End Function

Private Function Guid2Str(ByRef tGuid As GUID) As String
    Dim lngRet As Long
    Dim strGUID As String
    strGUID = String(40, 0)
    lngRet = StringFromGUID(tGuid, StrPtr(strGUID), Len(strGUID))
    If lngRet > 0 Then
        strGUID = Left$(strGUID, lngRet - 1)
    Else
        strGUID = vbNullString
    End If
    Guid2Str = strGUID
End Function

Private Function Ptr2Str(ByVal lpStr As Long) As String
    Dim lngLen As Long
    Dim bytBuffer() As Byte
    lngLen = lstrlenW(lpStr) * 2
    If lngLen > 0 Then
        ReDim bytBuffer(lngLen - 1)
        Call CopyMemory(bytBuffer(0), ByVal lpStr, lngLen)
        Call CoTaskMemFree(lpStr)
        Ptr2Str = bytBuffer
    End If
End Function

